import ast
import re
import base64
import json
import platform

import time
from threading import Thread

from maya import utils as mu
from ngSkinTools2 import cleanup, options, signal
from ngSkinTools2.api import plugin
from ngSkinTools2.log import getLogger
from ngSkinTools2.python_compatibility import Object
from ngSkinTools2.ui.parallel import ParallelTask

log = getLogger("license client")

try:
    from urllib2 import urlopen, Request, HTTPError
except:
    from urllib.request import urlopen, Request
    from urllib.error import HTTPError


# noinspection PyClassHasNoInit
class Status:
    ok = 0
    unknown = 1
    communicationError = 2
    invalidSignature = 100
    invalidHostId = 101

    @classmethod
    def asStr(cls, status):
        return {
            Status.ok: "",
            Status.unknown: "unknown error",
            Status.communicationError: "communication error",
            Status.invalidSignature: "invalid signature",
            Status.invalidHostId: "license is bound to another workstation",
        }[status]


class LicenseServerClient(Object):
    """
    this client helps c++ plugin communicate with the license server.
    this only takes care of transport layer, to keep options a bit more open for future like TLS communication,
    without having to embed a proper HTTP client into plugin
    """

    def __init__(self, timeout=30):
        self.serverUrl = "http://127.0.0.1:9050"
        self.sleepPeriod = 60*2
        self.timeout = timeout
        # indirectly referencing a function to call main thread to be able to have different behavior in tests
        self.mainThreadFunc = mu.executeInMainThreadWithResult
        self.lastError = None

        self.updateThreadRunning = False

    def refresh_reservation(self):
        """
        Called periodically to synchronize with license server
        :return:
        """

        req = Request(self.serverUrl + "/seat-reservations")

        def checkout_request():
            return plugin.ngst2License(serverRequest=True, hostName=platform.node())

        def checkin_response(resp):
            return plugin.ngst2License(serverResponse=resp)

        reqContents = self.mainThreadFunc(checkout_request)

        req.data = reqContents.encode()
        req.add_header("content-type", "application/json")

        try:
            resp = urlopen(req, timeout=self.timeout).read()
            self.lastError = None
            self.mainThreadFunc(checkin_response, resp)
        except IOError as err:
            log.error("ngSkinTools: error communicating with license server ({0}): {1} ".format(self.serverUrl, err))
            self.lastError = str(err)
            return


    def run_license_reservation_thread(self):
        if self.updateThreadRunning:
            return
        self.updateThreadRunning = True

        def threadFunc():
            while self.updateThreadRunning:
                self.refresh_reservation()
                time.sleep(self.sleepPeriod)

        def stop():
            self.updateThreadRunning = False

        Thread(target=threadFunc).start()
        cleanup.registerCleanupHandler(stop)


class LicenseFileHandler:
    def __init__(self):
        self.licenseServer = 'https://licensing.ngskintools.com/api/projects/ngskintools2/licenses/'
        self.changed = signal.Signal("license file changed")

    def get_host_id(self):
        result = plugin.ngst2License(q=True, hostid=True)
        return result

    def __parse_license_contents(self, contents):
        '''

        Example valid contents:

        LICENSE ngstkey ngskintools 1 standalone hostid=123-123
            sig=363c8f7f19679efc324b5ec713ffcf8968b3f4741b3b0d64f62436f8de799ec5

        For historical reasons, this is not a JSON.

        :param str contents:
        '''

        values = contents.split()
        if len(values) < 6:
            return None

        result = {}
        for index, v in enumerate(['stamp', 'file-type', 'product', 'product-version', 'license-type']):
            result[v] = values[index]

        # check what we have here
        if result['stamp'] != 'LICENSE':
            return None

        # check if license type is ngstkey
        if result['file-type'] != 'ngstkey':
            return None

        for value in values[5:]:
            k, v = value.split("=", 2)
            result[k] = v

        # signature is required
        if result.get('sig', None) is None:
            return None

        if result.get('hostid', None) is None:
            return None

        return result

    def apply_license_file(self, license_file):
        parsed_license_contents = self.__parse_license_contents(license_file)
        if parsed_license_contents is None:
            raise Exception

        conf = Configuration()
        conf.host_id = parsed_license_contents['hostid']
        conf.license_key = parsed_license_contents.get('licensekey', "")
        conf.signature = parsed_license_contents['sig']
        conf.type = Configuration.type_host_id

        self.apply_configuration(conf)
        log.info("saving configuration")
        conf.save()

    def apply_configuration(self, conf):
        if conf.type == Configuration.type_host_id:
            log.info("applying configuration: %s", json.dumps(conf.__dict__, indent=4))
            plugin.ngst2License(validateLicense=True, hostid=conf.host_id, licenseKey=conf.license_key,
                                signature=conf.signature)
            self.changed.emit()

    def download_license_file(self, license_key, host_id):
        '''
        exchanges licenseKey+hostId for a license file online.
        :type license_key: basestring
        :type host_id: basestring

        '''

        host_id = host_id.strip()
        if host_id == "":
            raise Exception("Cannot download license when host ID is empty")

        import json
        try:
            req = Request(self.licenseServer + license_key, headers={
                'Accept': 'application/vnd.releasedb.v1.1+json',
            })
            resp = urlopen(req, json.dumps({"hostId": host_id, "licenseFileType": 'ngstKey'}).encode())
            result = json.load(resp)
            return result['licenseFile']
        except HTTPError as err:
            code = err.getcode()

            message = str(err)
            try:
                message = json.load(err)['message']
            except:
                pass

            log.info("received error: %r", message)

            # for client-side errors, just convey the error from the server
            if code == 400:
                raise Exception(message)

            if code == 404:
                raise Exception("invalid license key")

            raise Exception("Failed downloading license information ({0}): {1}".format(err.getcode(), message))
        except Exception as err:
            raise Exception("Failed downloading license information: unknown error ({0})".format(str(err)))

    def create_online_key_download_task(self, license_key):
        """
        :param basestring license_key:
        """
        hostId = self.get_host_id()

        def run(context):
            context.error = ""
            context.license_file = None

            try:
                context.license_file = self.download_license_file(license_key, hostId)
            except Exception as err:
                context.error = err.message

        def done(context):
            if context.license_file is not None:
                self.apply_license_file(context.license_file)


        task = ParallelTask()
        task.add_run_handler(run)
        task.add_done_handler(done)

        return task

    def __is_license_key_valid(self, license_key):
        # UUID format:  8-4-4-4-12 hexadecimal strings
        # e.g. 12345678-abcd-ef09-1234-56789abcdef0
        format = '[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}'
        return re.match(format, license_key)


    def generate_activation_code_link(self, license_key):
        # sample key
        if not self.__is_license_key_valid(license_key):
            raise Exception("Invalid license key format")

        return "https://licensing.ngskintools.com/#/self-service/" + base64.b64encode(json.dumps({
            'type': 'activationCode',
            'product': 'ngskintools2',
            'licenseKey': license_key,
            'hostId': self.get_host_id()
        }).encode()).decode()

    def accept_activation_code(self, activation_code):
        try:
            contents = json.loads(base64.b64decode(activation_code))
        except:
            raise Exception("Invalid activation code: could not parse contents")

        self.apply_license_file(contents['licenseFile'])



class Configuration:
    CONFIGURATION_VAR = options.VAR_OPTION_PREFIX + 'licenseConfig'
    CONFIGURATION_ENV_VAR = 'NGSKINTOOLS2_LICENSE_CONFIG'

    type_host_id = "host-id"
    type_license_server = "license-server"

    json_mapping = {
        'type': 'type',
        'license_key': 'license-key',
        'host_id': 'host-id',
        'signature': 'signature',
    }

    def __init__(self):
        self.type = None
        self.license_key = ""
        self.host_id = ""
        self.signature = ""

        self.is_editable = True

    @classmethod
    def load(cls):
        import os
        data = os.getenv(cls.CONFIGURATION_ENV_VAR, None)

        result = Configuration()
        if data:
            result.is_editable = False # can't edit if configuration comes from environment variable
        if not data:
            data = options.loadOption(cls.CONFIGURATION_VAR, "")

        result.load_from_string(data)
        return result

    def load_from_string(self, s):
        """
        :param basestring s: either a python dictionary literal or a json containing key:value configuration of this object
        """
        s = s.strip()
        if not s:
            return

        log.info("loading config from: '%s'", s)
        try:
            val = json.loads(s)
        except:
            # if this fails, just raise error
            val = ast.literal_eval(s)

        for k, v in self.json_mapping.items():
            if v in val:
                setattr(self, k, val[v])

    def save_to_string(self):
        """
        :return: json serialization of this object, suitable for load_from_string later
        """
        import json
        return json.dumps({v: getattr(self, k) for k, v in self.json_mapping.items()})

    def save(self):
        if self.is_editable:
            options.saveOption(self.CONFIGURATION_VAR, self.save_to_string())


class LicenseData:
    def __init__(self):
        self.active = False
        self.status_description = ""
        self.licensed_to = ''


class LicenseClient:

    def __init__(self):
        self.serverClient = LicenseServerClient()
        self.licenseFileHandler = LicenseFileHandler()

        self.apply_configuration(Configuration.load())
        self.statusChanged = signal.Signal("license status changed")

        signal.on(self.licenseFileHandler.changed)(self.statusChanged.emit)


    def current_status(self):
        data = LicenseData()
        status_code = plugin.ngst2License(q=True, licenseStatus=True)
        data.active = status_code == 0
        data.licensed_to = plugin.ngst2License(q=True, licensedTo=True)
        data.status_description = "License active" if data.active else \
            "License activation problem: " + Status.asStr(status_code)
        return data

    def activate_with_license_key(self, license_key):
        return self.licenseFileHandler.create_online_key_download_task(license_key=license_key)

    def apply_configuration(self, config):
        self.licenseFileHandler.apply_configuration(config)

    def clear_configuration(self):
        config = Configuration()
        self.apply_configuration(config)
        plugin.ngst2License(reset=True)
        config.save()
        self.statusChanged.emit()

    def generate_acivation_code_link(self, license_key):
        return self.licenseFileHandler.generate_activation_code_link(license_key)

    def accept_activation_code(self, activation_code):
        self.licenseFileHandler.accept_activation_code(activation_code)

    def accept_license_server_url(self, server_address):
        """
        :type server_address: basestring
        """
        server_address = server_address.strip()
        if not server_address.lower().startswith('http://') and not server_address.lower().startswith('https://'):
            server_address = 'http://'+server_address

        self.serverClient.serverUrl = server_address
        self.serverClient.refresh_reservation()
        self.serverClient.run_license_reservation_thread()



licenseClient = LicenseClient()

