import ngSkinTools2.api
from PySide2 import QtWidgets, QtCore
from ngSkinTools2.ui import qt, dialogs

from maya import cmds

from ngSkinTools2 import signal
from ngSkinTools2.decorators import undoable
from ngSkinTools2.log import getLogger
from ngSkinTools2.ui.session import session, withSession

logger = getLogger("layer operations")


@withSession
@undoable
def initializeLayers(createFirstLayer=True):
    if session.state.layersAvailable:
        return  # ignore


    target = session.state.selectedSkinCluster

    layers = ngSkinTools2.api.init_layers(target)
    with ngSkinTools2.api.suspend_updates(target):
        if createFirstLayer:
            layer = layers.add("Base weights")
            layer.set_current()

    session.events.targetChanged.emitIfChanged()

    if ngSkinTools2.api.is_slow_mode_skin_cluster(target):
        dialogs.info("ngSkinTools switched to slow maya API for setting skin cluster weights for this skinCluster, to workaround a Maya bug when skinCluster uses dg nodes as inputs")


@undoable
def addLayer():
    layers = ngSkinTools2.api.Layers(session.state.selectedSkinCluster)

    def guessParent():
        currentLayer = layers.current_layer()
        if currentLayer is None:
            return None

        # current layer is a parent?
        if currentLayer.num_children > 0:
            return currentLayer

        return currentLayer.parent

    def generateName():
        name = "New Layer"
        currentLayerNames = [i.name for i in session.state.layers]
        logger.info(repr(currentLayerNames))
        index = 1
        while name in currentLayerNames:
            index += 1
            name = "New Layer ({0})".format(index)

        logger.info("%r not found in %r", name, currentLayerNames)
        return name

    with ngSkinTools2.api.suspend_updates(layers.mesh):
        new_layer = layers.add(generateName())
        new_layer.parent = guessParent()

        setCurrentLayer(new_layer)

    session.events.layerListChanged.emitIfChanged()

    return new_layer


def build_action_initialize_layers(session, parent):
    from ngSkinTools2.ui import actions
    from . import import_v1_actions

    def do_initialize():
        if import_v1_actions.can_import(session):
            q = "Skinning layers from previous version of ngSkinTools are present on this mesh. This operation will initialize " \
                "skinning layers from scratch, discarding previous layers information. Do you want to continue?"
            if not dialogs.yesNo(q):
                return

        initializeLayers()

    result = actions.define_action(parent, "Initialize Skinning Layers", callback=do_initialize)

    @signal.on(session.events.nodeSelectionChanged)
    def update():
        result.setEnabled(session.state.selectedSkinCluster is not None)

    return result


def buildAction_createLayer(session, parent):
    from ngSkinTools2.ui import actions

    result = actions.define_action(
        parent, "Create Layer", callback=addLayer, icon=":/newLayerEmpty.png",
        shortcut=QtCore.Qt.Key_Insert)

    @signal.on(session.events.targetChanged)
    def update_to_target():
        result.setEnabled(session.state.layersAvailable)

    update_to_target()

    return result


def buildAction_deleteLayer(session, parent):
    from ngSkinTools2.ui import actions

    result = actions.define_action(
        parent, "Delete Layer", callback=deleteSelectedLayers,
        shortcut=QtCore.Qt.Key_Delete)

    @signal.on(session.context.selectedLayers.changed, session.events.targetChanged, qtParent=parent)
    def update_to_target():
        result.setEnabled(session.state.layersAvailable and bool(session.context.selectedLayers(default=[])))

    update_to_target()

    return result


@undoable
def setCurrentLayer(layer):
    """
    :type layer: ngSkinTools2.api.layers.Layer
    """
    if not session.active():
        logger.info("didn't set current layer: no session")

    if not session.state.layersAvailable:
        logger.info("didn't set current layer: layers not enabled")

    logger.info("setting current layer to %r on %r", layer, session.state.selectedSkinCluster)
    layer.set_current()
    session.events.currentLayerChanged.emitIfChanged()


def getCurrentLayer():
    layers = ngSkinTools2.api.Layers(session.state.selectedSkinCluster)
    return layers.current_layer()


@undoable
def renameLayer(layer, newName):
    layer.name = newName
    cmds.evalDeferred(session.events.layerListChanged.emitIfChanged)


@undoable
def deleteSelectedLayers():
    layers = ngSkinTools2.api.Layers(session.state.selectedSkinCluster)
    for i in session.context.selectedLayers(default=[]):
        layers.delete(i)

    session.events.layerListChanged.emitIfChanged()
    session.events.currentLayerChanged.emitIfChanged()


def buildAction_toggleEnabled(session, parent):
    '''
    builds a UI action for toggling enable/disable on selected layers
    :type session: ngSkinTools2.ui.session.Session
    '''

    result = QtWidgets.QAction("Enabled", parent)
    result.setCheckable(True)

    def isChecked():
        '''
        return true if most of selected layers are enabled
        :return:
        '''
        layers = session.context.selectedLayers(default=[])
        if not layers:
            return True

        majorityEnabled = 0
        for layer in layers:
            try:
                # eat up the exception if layer id is invalid
                majorityEnabled += 1 if layer.enabled else -1
            except:
                pass

        return majorityEnabled >= 0

    @signal.on(session.context.selectedLayers.changed, session.events.targetChanged, qtParent=parent)
    def update_action():
        result.setChecked(isChecked())
        result.setEnabled(session.state.layersAvailable and bool(session.context.selectedLayers(default=[])))

    @qt.on(result.triggered)
    def triggered():
        layers = ngSkinTools2.api.Layers(session.state.selectedSkinCluster)

        enabled = not isChecked()
        result.setChecked(enabled)
        selectedLayers = session.context.selectedLayers()
        if not selectedLayers:
            return

        for i in selectedLayers:
            i.enabled = enabled

        logger.info("layers toggled: %r", selectedLayers)

        session.events.layerListChanged.emitIfChanged()
        update_action()

    update_action()

    return result
