from ngSkinTools2.python_compatibility import Object
from functools import partial

from ngSkinTools2.log import getLogger
from ngSkinTools2 import cleanup

log = getLogger("signal")


# noinspection PyBroadException
class Signal(Object):
    """
    Signal class collects observers, interested in some particular event,and handles
    signaling them all when some event occurs. Both handling and signaling happens outside
    of signal's own code


    Handlers are processed breath first, in a queue based system.
    1. root signal fires, adds all it's handlers to the queue;
    2. queue starts being processed
    3. handlers fire more signals, in turn adding more handlers to the end of the queue.
    """

    all = []

    currentExecutionQueue = []

    def __init__(self, name):
        if name is None:
            raise Exception("need name for debug purposes later")
        self.name = name
        self.handlers = []
        self.executing = False

        self.reset()
        Signal.all.append(self)
        cleanup.registerCleanupHandler(self.reset)

    def reset(self):
        self.handlers = []
        self.executing = False

    def emitDeferred(self, *args):
        import maya.utils as mu
        mu.executeDeferred(self.emit, *args)

    @classmethod
    def processHandlerQueue(cls):
        """
        processes handlers in the order received.
        as handlers can produce more handlers, execution continues until the queue is empty
        :return:
        """

        safeguard = 100
        #log.info("handler queue started with %d items", len(cls.currentExecutionQueue))

        currentHandler = 0;
        queue = cls.currentExecutionQueue

        while currentHandler < len(queue):
            if currentHandler == safeguard:
                raise Exception("queue might be looped")

            try:
                queue[currentHandler]()
            except Exception:
                import ngSkinTools2
                if ngSkinTools2.DEBUG_MODE:
                    import sys
                    import traceback
                    traceback.print_exc(file=sys.__stderr__)

            currentHandler += 1;

        #log.info("handler queue finished with %d items", len(cls.currentExecutionQueue))
        cls.currentExecutionQueue = []

    def emit(self, *args):
        '''
        emit mostly just adds handlers to the processing queue,
        but if nobody is processing handlers at the emit time,
        it is started here as well.
        '''
        #log.info("emit: %s", self.name)
        if self.executing:
            raise Exception('Nested emit on %s detected' % self.name)

        shouldStartProcessing = len(Signal.currentExecutionQueue) == 0;

        for i in self.handlers[:]:
            Signal.currentExecutionQueue.append(partial(i, *args))

        # is this the first signal added? start processing
        if shouldStartProcessing:
            Signal.processHandlerQueue()

    def addHandler(self, handler, qtParent=None):
        if hasattr(handler, 'emit'):
            handler = handler.emit

        self.handlers.append(handler)

        def remove():
            return self.removeHandler(handler)

        if qtParent is not None:
            qtParent.destroyed.connect(remove)

        return remove

    def removeHandler(self, handler):
        try:
            self.handlers.remove(handler)
        except ValueError:
            # not found in list? no biggie.
            pass


def on(*signals, **kwargs):
    '''
    decorator for function: list signals that should fire for this function.

        @signal.on(signalReference)
        def something():
            ...
    '''

    def decorator(fn):
        for i in signals:
            i.addHandler(fn, **kwargs)
        return fn

    return decorator


class SignalGraphBuilder(Object):
    def __init__(self):
        self.eventList = []

    @classmethod
    def on(cls, *eventList):
        """
        :rtype: SignalGraphBuilder
        """

        result = SignalGraphBuilder()
        result.eventList = eventList
        return result

    def do(self, *handlerList):
        for e in self.eventList:
            for h in handlerList:
                e.addHandler(h)
