from PySide2 import QtCore, QtWidgets

import shiboken2

import maya.OpenMayaUI as omui
from maya import cmds

from .. import cleanup, hotkeys, version
from ..log import getLogger
from ..options import config
from . import tabPaint, targetui, tabMirror, tabTools, dialogs, tabLayerEffects, licensewindow, \
    aboutwindow, updatewindow, qt
from .session import session

log = getLogger("main window")


def getImagePath(fileName):
    import os
    for i in os.getenv("XBMLANGPATH", "").split(os.path.pathsep):
        result = os.path.join(i, fileName)
        if os.path.isfile(result):
            return result

    return fileName


def buildMenu(parent, actions):
    menu = QtWidgets.QMenuBar(parent=parent)

    sub = menu.addMenu("File")
    sub.setSeparatorsCollapsible(False)

    sep = sub.addSeparator().setText("Import/Export")

    sub.setTearOffEnabled(True)
    sub.addAction(actions.importFile)
    sub.addAction(actions.exportFile)

    sub = menu.addMenu("Layers")
    sub.setSeparatorsCollapsible(False)
    sub.addSeparator().setText("Layer actions")
    sub.setTearOffEnabled(True)
    sub.addAction(actions.initialize)
    sub.addAction(actions.import_v1)
    actions.addLayersActions(sub)
    sub.addSeparator().setText("Copy")
    sub.addAction(actions.transfer)

    sub = menu.addMenu("Tools")
    sub.setSeparatorsCollapsible(False)
    sub.setTearOffEnabled(True)
    sub.addAction(actions.toolsAssignFromClosestJoint)
    sub.addSeparator()
    sub.addAction(actions.transfer)
    sub.addSeparator()
    sub.addAction(actions.toolsDeleteCustomNodesOnSelection)
    sub.addAction(actions.toolsDeleteCustomNodes)

    sub = menu.addMenu("Help")
    sub.setTearOffEnabled(True)

    sub.addAction(actions.documentation.launch_status)
    sub.addSeparator()
    sub.addAction(actions.documentation.user_guide)
    sub.addAction(actions.documentation.api_root)
    sub.addAction(actions.documentation.changelog)
    sub.addAction(actions.documentation.contact)
    sub.addSeparator()
    sub.addAction("Register...").triggered.connect(lambda: licensewindow.show(parent))
    sub.addSeparator()
    sub.addAction(actions.check_for_updates)
    sub.addAction("About...").triggered.connect(lambda: aboutwindow.show(parent))

    return menu


def buildRmbMenuLayers(view, actions):
    view.setContextMenuPolicy(QtCore.Qt.ActionsContextMenu)
    actions.addLayersActions(view)


def buildUI(parent):
    """

    :type parent: QWidget
    """
    window = QtWidgets.QWidget(parent)

    session.addQtWidgetReference(window)

    from ngSkinTools2.ui.actions import Actions
    actions = Actions(parent=window, session=session)

    tabs = QtWidgets.QTabWidget(window)

    tabs.addTab(tabPaint.build_ui(tabs), "Paint")
    tabs.addTab(tabMirror.buildUI(tabs), "Mirror")
    tabs.addTab(tabLayerEffects.build_ui(tabs), "Effects")
    tabs.addTab(tabTools.buildUI(tabs, actions, session), "Tools")
    tabs.addTab(QtWidgets.QWidget(), "Options")
    tabs.setCurrentIndex(0)

    layersToolbar = QtWidgets.QToolBar()
    layersToolbar.addAction(actions.addLayer)
    layersToolbar.setOrientation(QtCore.Qt.Vertical)

    spacing_h = 5
    spacing_v = 5

    layers_row = targetui.buildTargetUI(window, actions, session)

    split = QtWidgets.QSplitter(orientation=QtCore.Qt.Vertical, parent=window)
    split.addWidget(layers_row)
    split.addWidget(tabs)
    split.setStretchFactor(0, 3)
    split.setStretchFactor(1, 2)
    split.setContentsMargins(spacing_h, spacing_v, spacing_h, spacing_v)

    layout = QtWidgets.QVBoxLayout(window)
    layout.setMargin(0)
    layout.addWidget(buildMenu(window, actions))
    layout.addWidget(split)

    window.setLayout(layout)

    hotkeys.installHotkeys()

    dialogs.promptsParent = window

    if config.checkForUpdatesAtStartup:
        updatewindow.silent_check_and_show_if_available(qt.mainWindow)

    return window


DOCK_NAME = 'ngSkinTools2_mainWindow'


def open():
    """
    opens main window
    """

    if not cmds.workspaceControl(DOCK_NAME, q=True, exists=True):
        log.info("creating workspace control")
        cmds.workspaceControl(DOCK_NAME,
                              retain=False,
                              floating=True,
                              # ttc=["AttributeEditor",-1],
                              uiScript="from ngSkinTools2.ui import mainwindow; mainwindow.resumeInWorkspaceControl()")

    # bring tab to front;
    cmds.evalDeferred(lambda *args: cmds.workspaceControl(DOCK_NAME, e=True, r=True));

    def close():
        from maya import cmds
        # noinspection PyBroadException
        try:
            cmds.deleteUI(DOCK_NAME)
        except:
            pass
        pass

    cleanup.registerCleanupHandler(close)


def resumeInWorkspaceControl():
    """
    this method is responsible for resuming workspace control when Maya is building/restoring UI as part of it's
    workspace management cycle (open UI for the first time, restart maya, change workspace, etc)
    """

    log.info("creating workspace control")

    cmds.workspaceControl(DOCK_NAME, e=True, label="ngSkinTools " + version.pluginVersion())
    widget = shiboken2.wrapInstance(int(omui.MQtUtil.findControl(DOCK_NAME)), QtWidgets.QWidget)

    ui = buildUI(widget)
    widget.layout().addWidget(ui)
