"""
UI session is running as long as any of the ngSkinTools UI windows are open.

"""
from ngSkinTools2.python_compatibility import Object
import functools

from ngSkinTools2 import cleanup, signal
from ngSkinTools2.api import mirror
from ngSkinTools2.ui import events
from ngSkinTools2.licenseClient import licenseClient
from ngSkinTools2.log import getLogger
from ngSkinTools2.observableValue import ObservableValue
from ngSkinTools2.ui import dialogs

log = getLogger("events")


def loadPlugin():
    from maya import cmds

    pluginBinary = 'ngSkinTools2'

    def isPluginLoaded():
        return cmds.pluginInfo(pluginBinary, q=True, loaded=True)

    if not isPluginLoaded():
        result = cmds.loadPlugin(pluginBinary, quiet=True)

    if not isPluginLoaded():
        dialogs.displayError("Failed to load the plugin. This is often a case-by-case issue - contact support.")
        return

    from ngSkinTools2 import version

    expectedVersion = version.pluginVersion()
    actualVersion = cmds.pluginInfo(pluginBinary, q=True, version=True)
    if actualVersion != expectedVersion:
        dialogs.displayError(
            "Invalid plugin version detected: required '{expectedVersion}', "
            "but was '{actualVersion}'. Clean reinstall recommended.".format(
                expectedVersion=expectedVersion, actualVersion=actualVersion))


class CurrentLayerState(Object):
    def __init__(self):
        self.selectedSkinCluster = None

        # will be None, when no current layer is available
        self.layer = None  # type: ngSkinTools2.api.Layer


class CurrentPaintTargetState(Object):
    def __init__(self):
        self.skinCluster = None
        self.layerId = None
        self.target = None


class State(Object):

    def __init__(self):
        self.layersAvailable = False
        self.selection = []
        self.selectedSkinCluster = None

        self.layers = []  # type: List[ngSkinTools2.api.Layer]
        self.currentLayer = CurrentLayerState()
        self.currentInfluence = CurrentPaintTargetState()

    def mirror(self):
        # type: () -> mirror.Mirror
        return mirror.Mirror(self.selectedSkinCluster)


class Context(Object):
    def __init__(self):
        self.selectedInfluences = ObservableValue()  # [] logicalIndex
        self.selectedLayers = ObservableValue()  # [] layerId


class Session(Object):
    def __init__(self):
        # reference objects that are keeping the session
        self.references = set()
        self.state = None  # type: State
        self.events = None  # type: events.Events
        self.context = Context()

        self.referenceId = 0

    def active(self):
        return len(self.references) > 0

    def start(self):
        log.info("STARTING SESSION")

        loadPlugin()

        self.state = State()
        self.events = events.Events(self.state)

        @signal.on(self.events.targetChanged)
        def on_target_change():
            log.info("clearing target context")
            self.context.selectedLayers.set([])
            self.context.selectedInfluences.set([])
        # licenseClient.runLicenseUpdateThread()

        self.events.nodeSelectionChanged.emit()

    def end(self):
        log.info("ENDING SESSION")
        cleanup.cleanup()
        self.state = None
        self.events = None
        pass

    def addReference(self):
        '''
        returns unique ID for this added reference; this value needs to be passed into removeReference();
        this ensures that reference holder does not remove other references rather than his own.
        :return:
        '''
        self.referenceId += 1
        if not self.active():
            self.start()

        self.references.add(self.referenceId)
        return self.referenceId

    def removeReference(self, referenceId):

        try:
            self.references.remove(referenceId)
        except KeyError:
            pass

        if not self.active():
            self.end()

    def addQtWidgetReference(self, widget):
        ref = self.addReference()

        widget.destroyed.connect(lambda: self.removeReference(ref))

    def reference(self):
        class Context(Object):
            def __init__(self, session):
                """
                :type session: Session
                :type refObj: object
                """

                self.session = session

            def __enter__(self):
                self.ref = self.session.addReference()

            def __exit__(self, exc_type, exc_val, exc_tb):
                self.session.removeReference(self.ref)

        return Context(self)


session = Session()


def withSession(func):
    '''
    decorator makes sure that single session is running throughout function's lifetime
    '''

    @functools.wraps(func)
    def result(*args, **kwargs):
        ref = session.addReference()
        try:
            return func(*args, **kwargs)
        finally:
            session.removeReference(ref)

    return result
