from PySide2 import QtWidgets
from ngSkinTools2 import signal
from ngSkinTools2.log import getLogger
from ngSkinTools2.api import MirrorOptions, VertexTransferMode
from ngSkinTools2.ui import qt
from ngSkinTools2.ui.layout import TabSetup, createTitledRow
from ngSkinTools2.ui.session import session
from ngSkinTools2.ui.widgets import NumberSliderGroup

log = getLogger("tab paint")


def buildUI(parentWindow):

    def buildMirroringOptionsGroup(mirrorOptions):
        """
        :type mirrorOptions: mirror.MirrorOptions
        """
        def getMirrorDirection():
            mirrorDirection = QtWidgets.QComboBox()
            mirrorDirection.addItem("Guess from stroke", MirrorOptions.directionGuess)
            mirrorDirection.addItem("Positive to negative", MirrorOptions.directionPositiveToNegative)
            mirrorDirection.addItem("Negative to positive", MirrorOptions.directionNegativeToPositive)
            mirrorDirection.addItem("Flip", MirrorOptions.directionFlip)
            mirrorDirection.setMinimumWidth(1)
            @qt.on(mirrorDirection.currentIndexChanged)
            def valueChanged():
                mirrorOptions.direction = mirrorDirection.currentData()
                log.info("new direction: %r", mirrorOptions.direction)

            return mirrorDirection

        def axis():
            mirrorAxis = QtWidgets.QComboBox()
            mirrorAxis.addItem("X", 'x')
            mirrorAxis.addItem("Y", 'y')
            mirrorAxis.addItem("Z", 'z')

            @qt.on(mirrorAxis.currentIndexChanged)
            def valueChanged():
                session.state.mirror().set_axis(mirrorAxis.currentData())

            @signal.on(session.events.targetChanged)
            def targetChanged():
                if session.state.layersAvailable:
                    mirrorAxis.setCurrentIndex(mirrorAxis.findData(session.state.mirror().axis()))

            targetChanged()

            return mirrorAxis

        def mirrorSeamWidth():
            seam_width_ctrl = NumberSliderGroup(maximum=100)

            @signal.on(seam_width_ctrl.valueChanged)
            def valueChanged():
                session.state.mirror().set_seam_width(seam_width_ctrl.value())

            @signal.on(session.events.targetChanged)
            def update_values():
                if session.state.layersAvailable:
                    seam_width_ctrl.setValue(session.state.mirror().seam_width())

            update_values()

            return seam_width_ctrl.layout()

        def elements():
            result = QtWidgets.QVBoxLayout()

            influences = QtWidgets.QCheckBox("Influence weights")
            influences.setChecked(mirrorOptions.mirrorWeights)

            mask = QtWidgets.QCheckBox("Layer mask")
            mask.setChecked(mirrorOptions.mirrorMask)

            dq = QtWidgets.QCheckBox("Dual quaternion weights")
            dq.setChecked(mirrorOptions.mirrorDq)

            for i in [influences, mask, dq]:
                result.addWidget(i)

            @qt.on(influences.toggled, mask.toggled, dq.toggled)
            def updatePref():
                mirrorOptions.mirrorDq = dq.isChecked()
                mirrorOptions.mirrorMask = mask.isChecked()
                mirrorOptions.mirrorWeights = influences.isChecked()

            return result

        result = QtWidgets.QGroupBox("Mirroring options")
        layout = QtWidgets.QVBoxLayout()
        result.setLayout(layout)
        layout.addLayout(createTitledRow("Axis:", axis()))
        layout.addLayout(createTitledRow("Direction:", getMirrorDirection()))
        layout.addLayout(createTitledRow("Seam width:", mirrorSeamWidth()))
        layout.addLayout(createTitledRow("Elements to mirror:", elements()))

        return result

    def vertexMappingGroup():
        vertexMappingMode = QtWidgets.QComboBox()
        vertexMappingMode.addItem("Closest point on surface", VertexTransferMode.closestPoint)
        vertexMappingMode.addItem("UV space", VertexTransferMode.uvSpace)

        result = QtWidgets.QGroupBox("Vertex Mapping")
        layout = QtWidgets.QVBoxLayout()
        layout.addLayout(createTitledRow("Mapping mode:", vertexMappingMode))
        result.setLayout(layout)

        @qt.on(vertexMappingMode.currentIndexChanged)
        def valueChanged():
            session.state.mirror().set_vertex_transfer_mode(vertexMappingMode.currentData())

        @signal.on(session.events.targetChanged)
        def targetChanged():
            if session.state.layersAvailable:
                vertexMappingMode.setCurrentIndex(vertexMappingMode.findData(session.state.mirror().vertex_transfer_mode()))

        return result

    def influenceMappingGroup():


        def editMapping():
            mapping = QtWidgets.QPushButton("Preview and edit mapping")

            singleWindowPolicy = qt.SingleWindowPolicy()

            @qt.on(mapping.clicked)
            def edit():
                from ngSkinTools2.ui import influenceMappingUI
                window = influenceMappingUI.open_ui_for_mesh(parentWindow, session.state.selectedSkinCluster)
                singleWindowPolicy.setCurrent(window)

            return mapping

        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(editMapping())

        result = QtWidgets.QGroupBox("Influences mapping")
        result.setLayout(layout)
        return result

    mirrorOptions = MirrorOptions()

    tab = TabSetup()
    tab.innerLayout.addWidget(buildMirroringOptionsGroup(mirrorOptions))
    tab.innerLayout.addWidget(vertexMappingGroup())
    tab.innerLayout.addWidget(influenceMappingGroup())
    tab.innerLayout.addStretch()

    btnMirror = QtWidgets.QPushButton("Mirror")
    tab.lowerButtonsRow.addWidget(btnMirror)
    @qt.on(btnMirror.clicked)
    def mirrorClicked():
        session.state.mirror().mirror(mirrorOptions)

    @signal.on(session.events.targetChanged, qtParent=tab.tabContents)
    def updateTabEnabled():
        tab.tabContents.setEnabled(session.state.layersAvailable)

    updateTabEnabled()

    return tab.tabContents
