from PySide2 import QtGui
from PySide2 import QtWidgets
from ngSkinTools2 import signal
from ngSkinTools2.api.paint import BrushProjectionMode
from ngSkinTools2.log import getLogger
from ngSkinTools2.api import PaintMode, BrushShape, PaintTool
from ngSkinTools2.ui import qt, widgets
from ngSkinTools2.ui.layout import createTitledRow, TabSetup
from ngSkinTools2.ui.session import session

log = getLogger("tab paint")


def build_ui(parent):
    paint = PaintTool()

    def build_brush_settings_group():
        def brush_mode_row3():
            row = QtWidgets.QVBoxLayout()

            group = QtWidgets.QActionGroup(parent)

            actions = {}

            def create_brush_mode_button(t, mode, label, tooltip):
                a = QtWidgets.QAction(label, parent)
                a.setToolTip(tooltip)
                a.setCheckable(True)
                actions[mode] = a
                group.addAction(a)

                @qt.on(a.toggled)
                def toggled(checked):
                    if checked and paint.paint_mode != mode:
                        paint.paint_mode = mode
                        session.events.tool_settings_changed.emit()

                t.addAction(a)

            t = QtWidgets.QToolBar()
            create_brush_mode_button(t, PaintMode.replace, "Replace", "Whatever")
            create_brush_mode_button(t, PaintMode.add, "Add", "")
            create_brush_mode_button(t, PaintMode.scale, "Scale", "")
            row.addWidget(t)

            t = QtWidgets.QToolBar()
            create_brush_mode_button(t, PaintMode.smooth, "Smooth", "")
            create_brush_mode_button(t, PaintMode.sharpen, "Sharpen", "")
            row.addWidget(t)

            @signal.on(session.events.tool_settings_changed, qtParent=row)
            def updateCurrentBrushMode():
                actions[paint.paint_mode].setChecked(True)

            updateCurrentBrushMode()

            return row

        def brushShapeRow():
            result = QtWidgets.QToolBar()
            group = QtWidgets.QActionGroup(parent)

            def addBrushShapeAction(icon, title, shape, checked=False):
                a = QtWidgets.QAction(title, parent)
                a.setCheckable(True)
                a.setIcon(QtGui.QIcon(icon))
                a.setChecked(checked)
                result.addAction(a)
                group.addAction(a)

                @qt.on(a.toggled)
                def toggled(checked):
                    if checked:
                        paint.brush_shape = shape
                        updateToolSettingsInUI()

                @signal.on(session.events.tool_settings_changed, qtParent=a)
                def tool_settings_changed():
                    a.setChecked(paint.brush_shape == shape)

            addBrushShapeAction(':/circleSolid.png', 'Solid', BrushShape.solid, checked=True)
            addBrushShapeAction(':/circlePoly.png', 'Smooth', BrushShape.smooth)
            addBrushShapeAction(':/circleGaus.png', 'Gaus', BrushShape.gaus)

            return result

        def brushProjectionModeRow():
            result = QtWidgets.QToolBar()
            group = QtWidgets.QActionGroup(parent)

            def add(icon, title, mode, checked=False):
                a = QtWidgets.QAction(title, parent)
                a.setCheckable(True)
                a.setIcon(QtGui.QIcon(icon))
                a.setChecked(checked)
                result.addAction(a)
                group.addAction(a)

                @qt.on(a.toggled)
                def toggled(checked):
                    if checked:
                        paint.brush_projection_mode = mode
                        updateToolSettingsInUI()

                @signal.on(session.events.tool_settings_changed, qtParent=a)
                def tool_settings_changed():
                    a.setChecked(paint.brush_projection_mode == mode)

            add(None, 'Surface', BrushProjectionMode.surface, checked=True)
            add(None, 'Screen', BrushProjectionMode.screen, checked=False)
            return result

        def stylusPressureSelection():
            result = QtWidgets.QComboBox()
            result.addItem("Unused")
            result.addItem("Multiply intensity")
            result.addItem("Multiply opacity")
            result.addItem("Multiply radius")
            return result

        layout = QtWidgets.QVBoxLayout()
        layout.addLayout(createTitledRow("Brush projection:", brushProjectionModeRow()))
        layout.addLayout(createTitledRow("Brush mode:", brush_mode_row3()))
        layout.addLayout(createTitledRow("Brush shape:", brushShapeRow()))
        intensity = widgets.NumberSliderGroup()
        radius = widgets.NumberSliderGroup(maximum=100,
                                           tooltip="You can also set brush radius by just holding <b>B</b> "
                                                   "and mouse-dragging in the viewport")
        iterations = widgets.NumberSliderGroup(type=int, minimum=1, maximum=100)
        layout.addLayout(createTitledRow("Intensity:", intensity.layout()))
        layout.addLayout(createTitledRow("Brush radius:", radius.layout()))
        layout.addLayout(createTitledRow("Brush iterations:", iterations.layout()))

        influencesLimit = widgets.NumberSliderGroup(type=int, minimum=0, maximum=10)
        layout.addLayout(createTitledRow("Influences limit:", influencesLimit.layout()))

        @signal.on(influencesLimit.valueChanged)
        def influencesLimitChanged():
            paint.influences_limit = influencesLimit.value()
            updateToolSettingsInUI()

        interactiveMirror = QtWidgets.QCheckBox("Interactive mirror")
        layout.addLayout(createTitledRow("", interactiveMirror))

        @qt.on(interactiveMirror.stateChanged)
        def interactiveMirrorChanged():
            paint.interactive_mirror = interactiveMirror.isChecked()
            updateToolSettingsInUI()

        sampleJointOnStrokeStart = QtWidgets.QCheckBox("Sample current joint on stroke start")
        layout.addLayout(createTitledRow("", sampleJointOnStrokeStart))

        @qt.on(sampleJointOnStrokeStart.stateChanged)
        def interactiveMirrorChanged():
            paint.sample_joint_on_stroke_start = sampleJointOnStrokeStart.isChecked()
            updateToolSettingsInUI()

        redistributeRemovedWeight = QtWidgets.QCheckBox("Distribute to other influences")
        layout.addLayout(createTitledRow("Removed weight:", redistributeRemovedWeight))

        @qt.on(redistributeRemovedWeight.stateChanged)
        def redistributeRemovedWeightChanged():
            paint.redistribute_removed_weight = redistributeRemovedWeight.isChecked()
            updateToolSettingsInUI()

        stylus = stylusPressureSelection()
        layout.addLayout(createTitledRow("Stylus pressure:", stylus))

        @signal.on(session.events.tool_settings_changed, qtParent=layout)
        def updateToolSettingsInUI():
            intensity.setValue(paint.brush_intensity)
            radius.setValue(paint.brush_radius)
            iterations.setValue(paint.brush_iterations)
            iterations.setEnabled(paint.paint_mode in [PaintMode.smooth, PaintMode.sharpen])
            stylus.setCurrentIndex(paint.tablet_mode)
            interactiveMirror.setChecked(paint.interactive_mirror)
            redistributeRemovedWeight.setChecked(paint.redistribute_removed_weight)
            influencesLimit.setValue(paint.influences_limit)
            sampleJointOnStrokeStart.setChecked(paint.sample_joint_on_stroke_start)

        @signal.on(radius.valueChanged, qtParent=layout)
        def radiusEdited():
            paint.brush_radius = radius.value()
            updateToolSettingsInUI()

        @signal.on(intensity.valueChanged, qtParent=layout)
        def intensityEdited():
            paint.brush_intensity = intensity.value()
            updateToolSettingsInUI()

        @signal.on(iterations.valueChanged, qtParent=layout)
        def iterationsEdited():
            paint.brush_iterations = iterations.value()
            updateToolSettingsInUI()

        @qt.on(stylus.currentIndexChanged)
        def stylusEdited():
            paint.tablet_mode = stylus.currentIndex()
            updateToolSettingsInUI()

        updateToolSettingsInUI()

        settingsGroup = QtWidgets.QGroupBox("Brush behavior")
        settingsGroup.setLayout(layout)
        return settingsGroup

    def buildDisplaySettings():
        result = QtWidgets.QGroupBox("Display settings")
        layout = QtWidgets.QVBoxLayout()

        influencesDisplay = QtWidgets.QComboBox()
        influencesDisplay.addItem("All influences, multiple colors")
        influencesDisplay.addItem("Current influence, grayscale")
        influencesDisplay.addItem("Current influence, colored")
        influencesDisplay.setMinimumWidth(1)
        influencesDisplay.setCurrentIndex(paint.weights_display_mode)

        @qt.on(influencesDisplay.currentIndexChanged)
        def influencesDisplayChanged():
            paint.weights_display_mode = influencesDisplay.currentIndex()

        layout.addLayout(createTitledRow("Influences display:", influencesDisplay))

        showEffects = QtWidgets.QCheckBox("Show layer effects")
        layout.addLayout(createTitledRow("", showEffects))

        showSelectedVertsOnly = QtWidgets.QCheckBox("Hide unselected vertices")
        layout.addLayout(createTitledRow("", showSelectedVertsOnly))

        @qt.on(showEffects.stateChanged)
        def showEffectsChanged():
            paint.layer_effects_display = showEffects.isChecked()

        @qt.on(showSelectedVertsOnly.stateChanged)
        def showSelectedVertsChanged():
            paint.show_selected_verts_only = showSelectedVertsOnly.isChecked()

        t = QtWidgets.QToolBar()
        toggleOriginalMesh = QtWidgets.QAction("Show Original Mesh", t)
        toggleOriginalMesh.setCheckable(True)

        @qt.on(toggleOriginalMesh.triggered)
        def toggleDisplayNodeVisible():
            paint.display_node_visible = not toggleOriginalMesh.isChecked()
            update_ui_to_tool()

        t.addAction(toggleOriginalMesh)

        @signal.on(session.events.toolChanged, qtParent=tab.tabContents)
        def update_ui_to_tool():
            toggleOriginalMesh.setChecked(paint.is_painting() and not paint.display_node_visible)

            showEffects.setChecked(paint.layer_effects_display)
            showSelectedVertsOnly.setChecked(paint.show_selected_verts_only)

        layout.addLayout(createTitledRow("", t))

        update_ui_to_tool()

        result.setLayout(layout)
        return result

    def paintButton():
        btn = QtWidgets.QPushButton("Paint")
        btn.setCheckable(True)

        @signal.on(session.events.toolChanged, qtParent=btn)
        def updatePaintEnabled():
            btn.setChecked(paint.is_painting())

        btn.clicked.connect(paint.start)

        updatePaintEnabled()
        return btn

    def floodButton():
        btn = QtWidgets.QPushButton("Flood")
        btn.clicked.connect(paint.flood)

        @signal.on(session.events.toolChanged, qtParent=btn)
        def updateEnabled():
            btn.setEnabled(paint.is_painting())

        updateEnabled()

        return btn

    tab = TabSetup()
    tab.innerLayout.addWidget(build_brush_settings_group())
    tab.innerLayout.addWidget(buildDisplaySettings())
    tab.innerLayout.addStretch()

    tab.lowerButtonsRow.addWidget(paintButton())
    tab.lowerButtonsRow.addWidget(floodButton())

    @signal.on(session.events.toolChanged, qtParent=tab.tabContents)
    def updateToTool():
        tab.scrollArea.setEnabled(paint.is_painting())

    @signal.on(session.events.targetChanged, qtParent=tab.tabContents)
    def updateTabEnabled():
        tab.tabContents.setEnabled(session.state.layersAvailable)

    updateToTool()
    updateTabEnabled()

    return tab.tabContents
